"""
docsig._main
============

Contains package entry point.
"""
from ._config import Parser as _Parser
from ._core import docsig as _docsig


def main() -> int:
    """Main function for package.

    Collect config and arguments for the commandline.

    :return: Exit status for whether test failed or not.
    """
    parser = _Parser()
    return _docsig(
        *parser.args.path,
        string=parser.args.string,
        check_class=parser.args.check_class,
        check_dunders=parser.args.check_dunders,
        check_overridden=parser.args.check_overridden,
        check_protected=parser.args.check_protected,
        no_ansi=parser.args.no_ansi,
        summary=parser.args.summary,
        targets=parser.args.target,
        disable=parser.args.disable
    )



"""
docsig._display
===============
"""
from __future__ import annotations

import typing as _t
from collections import UserString as _UserString

from object_colors import Color as _Color
from pygments import highlight as _highlight
from pygments.formatters.terminal256 import (
    Terminal256Formatter as _Terminal256Formatter,
)

# noinspection PyUnresolvedReferences
from pygments.lexers.python import PythonLexer as _PythonLexer

from ._function import ARG as _ARG
from ._function import KEY as _KEY
from ._function import Function as _Function
from ._function import Param as _Param
from ._objects import MutableMapping as _MutableMapping
from ._objects import MutableSequence as _MutableSequence
from ._report import Report as _Report

color = _Color()

color.populate_colors()

TAB = "    "


class _ANSI:
    def __init__(self, no_ansi: bool = False) -> None:
        self._no_ansi = no_ansi

    def color(self, obj: _t.Any, color_obj: _Color) -> str:
        """Get string with selected color.

        :param obj: Any object, represented as ``__str__``.
        :param color_obj: Instantiated ``Color`` object.
        :return: Colored string or string as was supplied.
        """
        return str(obj) if self._no_ansi else color_obj.get(obj)

    def syntax(self, obj: _t.Any) -> str:
        """Get code representation with syntax highlighting.

        :param obj: Any object, represented as ``__str__``.
        :return: Colored string or string as was supplied.
        """
        return (
            str(obj)
            if self._no_ansi
            else _highlight(
                obj, _PythonLexer(), _Terminal256Formatter(style="monokai")
            ).strip()
        )


class FuncStr(_UserString):
    """String representation for function.

    :param func: Represents a function with signature and docstring
        parameters.
    :param no_ansi: Disable ANSI output.
    """

    CHECK = "\u2713"
    CROSS = "\u2716"
    TRIPLE_QUOTES = '"""'

    def __init__(self, func: _Function, no_ansi: bool = False) -> None:
        super().__init__(func.name)
        self._ansi = _ANSI(no_ansi)
        self._parent_name = func.parent.name
        self._isinit = func.isinit
        self.data = ""
        self._is_string = func.docstring.string is not None
        if self._isinit:
            self.data += TAB

        self.data += self._ansi.syntax(f"def {func.name}(")
        if self._is_string:
            self._docstring = self._ansi.syntax(f"{TAB}{self.TRIPLE_QUOTES}")
        else:
            self._docstring = f"{TAB}{self._ansi.color('...', color.red)}\n"

        self._mark = self._ansi.color(self.CHECK, color.green)
        for index in range(len(func)):
            arg = func.signature.args.get(index)
            doc = func.docstring.args.get(index)
            self.add_param(arg, doc, arg != doc)
            if index + 1 != len(func):
                self.add_comma()

        self.set_mark()
        if func.docstring.returns and func.signature.returns:
            self.add_return()
        elif (
            func.docstring.returns
            and not func.signature.returns
            or func.signature.returns
            and not func.docstring.returns
        ):
            self.add_return(failed=True)

        self.close_sig(func.signature.rettype)
        self.close_docstring()
        self.render()

    def _cat_docstring(self, string: str) -> None:
        if self._is_string:
            self._docstring += string

    def set_mark(self, failed: bool = False) -> None:
        """Set mark to a cross or a check.

        :param failed: Boolean to test that check failed.
        """
        self._mark = (
            self._ansi.color(self.CROSS, color.red)
            if failed
            else self._ansi.color(self.CHECK, color.green)
        )

    def add_param(
        self, sig: _Param, doc: _Param, failed: bool = False
    ) -> None:
        """Add parameters to docstring.

        :param sig: Signature argument.
        :param doc: Docstring argument.
        :param failed: Boolean to test that check failed.
        """
        self.set_mark(failed)
        sig_name = sig.name
        if sig.kind == _KEY:
            sig_name = f"**{sig_name}"

        if sig.kind == _ARG:
            sig_name = f"*{sig_name}"

        self.data += f"{self._mark}{sig_name}"
        doc_name = doc.name
        if doc.kind == _KEY:
            doc_name = "(**)"

        self._cat_docstring(f"\n{TAB}:{doc.kind} {doc_name}: {self._mark}")

    def add_return(self, failed: bool = False) -> None:
        """Add return statement to docstring.

        :param failed: Boolean to test that check failed.
        """
        self.set_mark(failed)
        self._cat_docstring(f"\n{TAB}:return: {self._mark}")

    def close_sig(self, arg: str | None) -> None:
        """Close function signature.

        :param arg: Signature argument.
        """
        if arg is not None:
            self.data += "{}{}{}{}".format(
                self._ansi.syntax(") -> "),
                self._mark,
                arg,
                self._ansi.syntax(":"),
            )
        else:
            self.data += "{}{}{}".format(
                self._ansi.syntax(")"),
                self._ansi.color("?", color.red),
                self._ansi.syntax(":"),
            )

    def add_comma(self) -> None:
        """Add comma between parenthesis."""
        self.data += self._ansi.syntax(", ")

    def close_docstring(self) -> None:
        """Close docstring."""
        self._cat_docstring(
            f"\n{TAB}{self._ansi.syntax(self.TRIPLE_QUOTES)}\n"
        )

    def render(self) -> None:
        """Render final string by adding docstring to function."""
        if self._isinit:
            self.data = (
                self._ansi.syntax(f"class {self._parent_name}:")
                + f"\n{self._docstring}"
                + f"\n{self.data}\n"
            )
        else:
            self.data += f"\n{self._docstring}"


class Failure(_t.NamedTuple):
    """Failed function data."""

    func: _Function
    func_str: FuncStr
    report: _Report


class Failures(_MutableSequence[Failure]):
    """Sequence of failed functions."""


class _DisplaySequence(_MutableMapping[str, _t.List[Failures]]):
    """Sequence for collection of report info.

    If an attempt is made to append a report to a list whose key does
    not exist then the key and its list value will be added first.
    """

    def __getitem__(self, key: str) -> list[Failures]:
        if key not in super().__iter__():
            super().__setitem__(key, [])

        return super().__getitem__(key)


class Display(_DisplaySequence):
    """Collect and display report.

    :param no_ansi: Disable ANSI output.
    """

    def __init__(self, no_ansi: bool = False) -> None:
        super().__init__()
        self._ansi = _ANSI(no_ansi)

    def report(self) -> None:
        """Display report if any checks have failed."""
        for key, value in self.items():
            for failures in value:
                for failure in failures:
                    header = f"{key}{failure.func.lineno}"
                    if failure.func.parent.name:
                        header += f" in {failure.func.parent.name}"

                    print(self._ansi.color(header, color.magenta))
                    print(len(header) * "-")
                    print(failure.func_str)
                    print(failure.report.get_report())

    def summarise(self) -> None:
        """Display report summary if any checks have failed."""
        for key, value in self.items():
            for failures in value:
                for failure in failures:
                    header = f"{key}{failure.func.lineno}"
                    if failure.func.parent.name:
                        header += " in {}.{}".format(
                            failure.func.parent.name, failure.func.name
                        )

                    print(
                        "{}\n\t{}".format(
                            self._ansi.color(header, color.magenta),
                            failure.report.get_report("\t").strip(),
                        )
                    )



"""
docsig._report
==============
"""
from __future__ import annotations

from . import messages as _messages
from ._function import RETURN as _RETURN
from ._function import Function as _Function
from ._function import Param as _Param
from ._objects import MutableSequence as _MutableSequence
from ._utils import almost_equal as _almost_equal

_MIN_MATCH = 0.8
_MAX_MATCH = 1.0


class _MessageSequence(_MutableSequence[str]):
    def __init__(
        self,
        targets: list[str] | None = None,
        disable: list[str] | None = None,
    ) -> None:
        super().__init__()
        self._disable = disable or []
        self._disabled = False
        self._resolve_targeted(targets or [])
        self._errors: list[str] = []

    def _resolve_targeted(self, targets: list[str]) -> None:
        errors = [
            i
            for i in dir(_messages)
            if not i.startswith("__") and i.startswith("E")
        ]
        if targets:
            for target in targets:
                errors.remove(target)

            self._disable.extend(errors)

    def _lock(self, value: str) -> None:
        # if the last code to be disabled was an error then all
        # following hints are disabled until a new error is evaluated
        if value.startswith("E"):
            self._disabled = False

        if value in self._disable:
            self._disabled = True

    def insert(self, index: int, value: str) -> None:
        self._lock(value)
        if value.startswith("E"):
            self._errors.append(value)

        message = getattr(_messages, value)
        if not self._disabled and not (
            value.startswith("E") and message in self
        ):
            super().insert(index, message)


class Report(_MessageSequence):
    """Compile and produce report.

    :param func: Function object.
    :param targets: List of errors to target.
    :param disable: List of errors to disable.
    """

    def __init__(
        self,
        func: _Function,
        targets: list[str] | None = None,
        disable: list[str] | None = None,
    ) -> None:
        super().__init__(targets, disable)
        self._func = func

    def order(self, sig: _Param, doc: _Param) -> None:
        """Test for documented parameters and their order.

        :param sig: Signature argument.
        :param doc: Docstring argument.
        """
        if any(sig.name == i.name for i in self._func.docstring.args) or any(
            doc.name == i.name for i in self._func.signature.args
        ):
            self.append("E101")

    def exists(self) -> None:
        """Test that non-existing parameter is not documented."""
        if len(self._func.docstring.args) > len(self._func.signature.args):
            self.append("E102")

    def missing_func_docstring(self) -> None:
        """Test that docstring is not missing from func."""
        if not self._func.isinit and self._func.docstring.string is None:
            self.append("E113")

    def missing_class_docstring(self) -> None:
        """Test that docstring is not missing from class."""
        if self._func.isinit and self._func.docstring.string is None:
            self.append("E114")

    def missing(self) -> None:
        """Test that parameter is not missing from documentation."""
        if len(self._func.signature.args) > len(self._func.docstring.args):
            self.append("E103")

    def duplicates(self) -> None:
        """Test that there are no duplicate parameters in docstring."""
        if self._func.docstring.args.duplicated:
            self.append("E106")

    def extra_return(self) -> None:
        """Check that return is not documented when there is none."""
        if (
            self._func.docstring.returns
            and self._func.signature.rettype == "None"
            and not self._func.isproperty
            and not self._func.isinit
        ):
            self.append("E104")

    def property_return(self) -> None:
        """Check that return is not documented for property."""
        if self._func.docstring.returns and self._func.isproperty:
            self.append("E108")
            self.append("H101")

    def return_not_typed(self) -> None:
        """Check that return is not documented when no type provided."""
        if (
            self._func.signature.rettype is None
            and not self._func.isproperty
            and not self._func.isinit
        ):
            self.append("E109")

    def missing_return(self) -> None:
        """Check that return is documented when func returns value."""
        if (
            self._func.signature.returns
            and not self._func.docstring.returns
            and not self._func.isproperty
            and not self._func.isinit
        ):
            self.append("E105")
            docstring = self._func.docstring.string
            if docstring is not None and _RETURN in docstring:
                self.append("H102")

    def incorrect(self, sig: _Param, doc: _Param) -> None:
        """Test that proper syntax is used when documenting parameters.

        :param sig: Signature argument.
        :param doc: Docstring argument.
        """
        if sig.name is None and doc.name is None:
            self.append("E107")

    def not_equal(self, sig: _Param, doc: _Param) -> None:
        """Final catch-all.

        Only applies if no other errors, including disabled, have been
        triggered

        :param sig: Signature argument.
        :param doc: Docstring argument.
        """
        if sig.name is not None and doc.name is not None and not self._errors:
            self.append("E110")

    def class_return(self) -> None:
        """Check that return is not documented for __init__."""
        if self._func.docstring.returns and self._func.isinit:
            self.append("E111")
            self.append("H103")

    def misspelled(self, sig: _Param, doc: _Param) -> None:
        """Test whether there is a spelling error in documentation.

        To avoid false positives also check whether doc param is almost
        equal amongst its sibling params. If params are too similarly
        named then this error won't be raised.

        :param sig: Signature argument.
        :param doc: Docstring argument.
        """
        if (
            sig.name is not None
            and doc.name is not None
            and not self._errors
            and _almost_equal(sig.name, doc.name, _MIN_MATCH, _MAX_MATCH)
        ):
            self.append("E112")

    def description_syntax(self, doc: _Param) -> None:
        """Test whether docstring description has correct spacing.

        :param doc: Docstring argument.
        """
        if doc.description is not None and not doc.description.startswith(" "):
            self.append("E115")

    def indent_syntax(self, doc: _Param) -> None:
        """Test whether docstring description is indented correctly.

        :param doc: Docstring argument.
        """
        if doc.indent > 0:
            self.append("E116")

    def get_report(self, prefix: str = "") -> str:
        """Get report compiled as a string.

        :param prefix: Prefix report.
        :return: Current report.
        """
        return f"\n{prefix}".join(self) + "\n"


def generate_report(
    func: _Function, targets: list[str], disable: list[str]
) -> Report:
    """Generate report if function or method has failed.

    :param func: Function object.
    :param targets: List of errors to target.
    :param disable: List of errors to disable.
    :return: Compiled report.
    """
    report = Report(func, targets, disable)
    report.missing_class_docstring()
    report.missing_func_docstring()
    report.return_not_typed()
    if func.docstring.string is not None:
        report.exists()
        report.missing()
        report.duplicates()
        report.extra_return()
        report.missing_return()
        report.property_return()
        report.class_return()
        for index in range(len(func)):
            arg = func.signature.args.get(index)
            doc = func.docstring.args.get(index)
            report.description_syntax(doc)
            report.indent_syntax(doc)
            if arg != doc:
                report.order(arg, doc)
                report.incorrect(arg, doc)
                report.misspelled(arg, doc)
                report.not_equal(arg, doc)

    return report



"""
docsig._function
================
"""
from __future__ import annotations

import re as _re
import textwrap as _textwrap
import typing as _t
from collections import Counter as _Counter

import astroid as _ast
import sphinxcontrib.napoleon as _s

from ._objects import MutableSequence as _MutableSequence
from ._utils import isprotected as _isprotected

PARAM = "param"
KEYWORD = "keyword"
KEY = "key"
RETURN = "return"
ARG = "arg"


class _GoogleDocstring(str):
    def __new__(cls, string: str) -> _GoogleDocstring:
        return super().__new__(cls, str(_s.GoogleDocstring(string)))


class _NumpyDocstring(str):
    def __new__(cls, string: str) -> _NumpyDocstring:
        return super().__new__(cls, str(_s.NumpyDocstring(string)))


class _DocFmt(str):
    def __new__(cls, string: str) -> _DocFmt:
        return super().__new__(
            cls,
            _textwrap.dedent("\n".join(string.splitlines()[1:])).replace(
                "*", ""
            ),
        )


class _RawDocstring(str):
    def __new__(cls, string: str) -> _RawDocstring:
        return super().__new__(
            cls, _NumpyDocstring(_GoogleDocstring(_DocFmt(string)))
        )


class Param(_t.NamedTuple):
    """A tuple of param types and their names."""

    kind: str = PARAM
    name: str | None = None
    description: str | None = None
    indent: int = 0

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Param):
            return False

        args = self, other
        return all(i.kind == KEY for i in args) or (
            self.name == other.name and all(i.name is not None for i in args)
        )

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def isprotected(self) -> bool:
        """Boolean value for whether parameter is protected."""
        return str(self.name).startswith("_")


class _Matches(_MutableSequence[Param]):
    _pattern = _re.compile(":(.*?):")
    _normalize = {KEYWORD: KEY}

    def __init__(self, string: str) -> None:
        super().__init__()
        for line in string.splitlines():
            strip_line = line.lstrip()
            match = self._pattern.split(strip_line)[1:]
            if match:
                name = description = None
                kinds = match[0].split()
                if kinds:
                    kind = kinds[0]
                    for substring, replace in self._normalize.items():
                        kind = kind.replace(substring, replace)

                    if len(kinds) > 1:
                        name = kinds[1]

                    if len(match) > 1:
                        description = match[1]

                    super().append(
                        Param(
                            kind,
                            name,
                            description,
                            len(line) - len(strip_line),
                        )
                    )


class _Params(_MutableSequence[Param]):
    def insert(self, index: int, value: Param) -> None:
        if not value.isprotected and (
            value.kind in (PARAM, ARG)
            or (value.kind == KEY and not any(i.kind == KEY for i in self))
        ):
            super().insert(index, value)

    def get(self, index: int) -> Param:
        """Get a param.

        If the index does not exist return a `Param` with None as
        `Param.name`.

        :param index: Index of param to get.
        :return: Param belonging to the index.
        """
        try:
            return self[index]
        except IndexError:
            return Param()

    @property
    def duplicated(self) -> bool:
        """Boolean value for whether there are duplicate parameters."""
        return any(k for k, v in _Counter(self).items() if v > 1)


class _DocSig:
    def __init__(self) -> None:
        self._args = _Params()
        self._returns = False

    @property
    def args(self) -> _Params:
        """Collection of `Param` types."""
        return self._args

    @property
    def returns(self) -> bool:
        """Boolean value for whether this returns a value."""
        return self._returns


class _Signature(_DocSig):
    def __init__(
        self,
        arguments: _ast.Arguments,
        returns: _ast.Module,
        ismethod: bool = False,
        isstaticmethod: bool = False,
    ) -> None:
        super().__init__()
        if ismethod and not isstaticmethod and arguments.args:
            arguments.args.pop(0)

        self._args.extend(Param(name=a.name) for a in arguments.args)
        if arguments.vararg is not None:
            self._args.append(Param(ARG, name=arguments.vararg))

        self._args.extend(Param(name=k.name) for k in arguments.kwonlyargs)
        if arguments.kwarg is not None:
            self._args.append(Param(KEY, name=arguments.kwarg))

        self._rettype = self._get_rettype(returns)
        self._returns = str(self._rettype) != "None"

    def _get_rettype(self, returns: _ast.NodeNG | None) -> str | None:
        if isinstance(returns, _ast.Name):
            return returns.name

        if isinstance(returns, _ast.Attribute):
            return returns.attrname

        if isinstance(returns, _ast.Const):
            return str(returns.kind)

        if isinstance(returns, _ast.Subscript):
            return "{}[{}]".format(
                self._get_rettype(returns.value),
                self._get_rettype(returns.slice),
            )

        if isinstance(returns, _ast.BinOp):
            return "{} | {}".format(
                self._get_rettype(returns.left),
                self._get_rettype(returns.right),
            )

        return None

    @property
    def rettype(self) -> str | None:
        """Function's return value.

        If a function is typed to return None, return str(None). If no
        typehint exists then return None (NoneType).
        """
        return self._rettype


class _Docstring(_DocSig):
    def __init__(self, node: _ast.Const | None = None) -> None:
        super().__init__()
        self._string = None
        if node is not None:
            self._string = _RawDocstring(node.value)
            self._args.extend(_Matches(self._string))

        self._returns = self._string is not None and bool(
            _re.search(":returns?:", self._string)
        )

    @property
    def string(self) -> _RawDocstring | None:
        """The raw documentation string, if it exists, else None."""
        return self._string


class Function:
    """Represents a function with signature and docstring parameters.

    :param node: Function's abstract syntax tree.
    """

    def __init__(self, node: _ast.FunctionDef) -> None:
        self._node = node
        self._parent = node.parent.frame()
        self._signature = _Signature(
            node.args, node.returns, self.ismethod, self.isstaticmethod
        )
        self._docstring = _Docstring(
            node.doc_node if not self.isinit else self._parent.doc_node
        )

    def __len__(self) -> int:
        """Length of the longest sequence of args."""
        return max([len(self.signature.args), len(self.docstring.args)])

    def _decorated_with(self, name: str) -> bool:
        if self._node.decorators is not None:
            for dec in self._node.decorators.nodes:
                return isinstance(dec, _ast.Name) and dec.name == name

        return False

    @property
    def ismethod(self) -> bool:
        """Boolean value for whether function is a method."""
        return isinstance(self._parent, _ast.ClassDef)

    @property
    def isproperty(self) -> bool:
        """Boolean value for whether function is a property."""
        return self.ismethod and self._decorated_with("property")

    @property
    def isinit(self) -> bool:
        """Boolean value for whether function is a class constructor."""
        return self.ismethod and self.name == "__init__"

    @property
    def isoverridden(self) -> bool:
        """Boolean value for whether function is overridden."""
        if self.ismethod and not self.isinit:
            for ancestor in self._parent.ancestors():
                if self.name in ancestor and isinstance(
                    ancestor[self.name], _ast.FunctionDef
                ):
                    return True

        return False

    @property
    def isprotected(self) -> bool:
        """Boolean value for whether function is protected."""
        return (
            _isprotected(self.name) and not self.isinit and not self.isdunder
        )

    @property
    def isstaticmethod(self) -> bool:
        """Boolean value for whether function is a static method."""
        return self.ismethod and self._decorated_with("staticmethod")

    @property
    def isdunder(self) -> bool:
        """Boolean value for whether function is a dunder method."""
        return (
            self.ismethod
            and not self.isinit
            and self.name[:2] + self.name[-2:] == "____"
        )

    @property
    def name(self) -> str:
        """The name of the function."""
        return self._node.name

    @property
    def parent(
        self,
    ) -> _ast.FunctionDef | _ast.Module | _ast.ClassDef | _ast.Lambda:
        """Function's parent node."""
        return self._parent

    @property
    def lineno(self) -> int:
        """Line number of function declaration."""
        return self._node.lineno or 0

    @property
    def signature(self) -> _Signature:
        """The function's signature parameters."""
        return self._signature

    @property
    def docstring(self) -> _Docstring:
        """The function's docstring."""
        return self._docstring



"""
docsig.messages
===============
| E101: parameters out of order
| E102: includes parameters that do not exist
| E103: parameters missing
| E104: return statement documented for None
| E105: return missing from docstring
| E106: duplicate parameters found
| E107: parameter appears to be incorrectly documented
| E108: return statement documented for property
| E109: cannot determine whether a return statement should exist or not
| E110: documented parameter not equal to its respective argument
| E111: return statement documented for class
| E112: spelling error found in documented parameter
| E113: function is missing a docstring
| E114: class is missing a docstring
| E115: syntax error in description
| E116: param not indented correctly
"""
# Exxx: Error
# E1xx: Docstring
E101 = "E101: parameters out of order"
E102 = "E102: includes parameters that do not exist"
E103 = "E103: parameters missing"
E104 = "E104: return statement documented for None"
E105 = "E105: return missing from docstring"
E106 = "E106: duplicate parameters found"
E107 = "E107: parameter appears to be incorrectly documented"
E108 = "E108: return statement documented for property"
E109 = "E109: cannot determine whether a return statement should exist or not"
E110 = "E110: documented parameter not equal to its respective argument"
E111 = "E111: return statement documented for class"
E112 = "E112: spelling error found in documented parameter"
E113 = "E113: function is missing a docstring"
E114 = "E114: class is missing a docstring"
E115 = "E115: syntax error in description"
E116 = "E116: param not indented correctly"

# Hxxx: Hint
# H1xx: Docstring
H101 = "hint: documentation is sufficient as a getter is the value returned"
H102 = "hint: it is possible a syntax error could be causing this"
H103 = "hint: a class does not return a value during instantiation"



"""Check signature params for proper documentation."""
from . import messages
from ._core import docsig
from ._main import main
from ._version import __version__

__all__ = ["__version__", "docsig", "main", "messages"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['docsig']

package_data = \
{'': ['*']}

install_requires = \
['Pygments>=2.13.0,<3.0.0',
 'arcon>=0.1.1,<0.2.0',
 'astroid>=2.11.6,<3.0.0',
 'object-colors>=2.1.0,<3.0.0',
 'sphinxcontrib-napoleon>=0.7,<0.8']

entry_points = \
{'console_scripts': ['docsig = docsig.__main__:main']}

setup_kwargs = {
    'name': 'docsig',
    'version': '0.29.0',
    'description': 'Check signature params for proper documentation',
    'long_description': 'docsig\n======\n.. image:: https://img.shields.io/badge/License-MIT-yellow.svg\n    :target: https://opensource.org/licenses/MIT\n    :alt: License\n.. image:: https://img.shields.io/pypi/v/docsig\n    :target: https://img.shields.io/pypi/v/docsig\n    :alt: pypi\n.. image:: https://github.com/jshwi/docsig/actions/workflows/ci.yml/badge.svg\n    :target: https://github.com/jshwi/docsig/actions/workflows/ci.yml\n    :alt: CI\n.. image:: https://codecov.io/gh/jshwi/docsig/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/jshwi/docsig\n    :alt: codecov.io\n.. image:: https://readthedocs.org/projects/docsig/badge/?version=latest\n    :target: https://docsig.readthedocs.io/en/latest/?badge=latest\n    :alt: readthedocs.org\n.. image:: https://img.shields.io/badge/python-3.8-blue.svg\n    :target: https://www.python.org/downloads/release/python-380\n    :alt: python3.8\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: black\n\nCheck signature params for proper documentation\n-----------------------------------------------\n\nSupports reStructuredText (Sphinx), `numpy`, and `Google`\n\nInstallation\n------------\n\n.. code-block:: console\n\n    $ pip install docsig\n\nUsage\n-----\n\nCommandline\n***********\n\n.. code-block:: console\n\n    usage: docsig [-h] [-v] [-c] [-D] [-o] [-p] [-n] [-S] [-s STR] [-d LIST] [-t LIST] [path [path ...]]\n\n    Check signature params for proper documentation\n\n    positional arguments:\n      path                     directories or files to check (default: .)\n\n    optional arguments:\n      -h, --help               show this help message and exit\n      -v, --version            show version and exit\n      -c, --check-class        check class docstrings\n      -D, --check-dunders      check dunder methods\n      -o, --check-overridden   check overridden methods\n      -p, --check-protected    check protected functions and classes\n      -n, --no-ansi            disable ansi output\n      -S, --summary            print a summarised report\n      -s STR, --string STR     string to parse instead of files\n      -d LIST, --disable LIST  comma separated list of rules to disable\n      -t LIST, --target LIST   comma separated list of rules to target\n\nOptions can also be configured with the pyproject.toml file\n\nIf you find the output is too verbose then the report can be configured to display a summary\n\n.. code-block:: toml\n\n    [tool.docsig]\n    check-dunders = false\n    check-overridden = false\n    check-protected = false\n    summary = true\n    disable = [\n        "E101",\n        "E102",\n        "E103",\n    ]\n    target = [\n        "E102",\n        "E103",\n        "E104",\n    ]\n\nAPI\n***\n\n.. code-block:: python\n\n    >>> from docsig import docsig\n\n.. code-block:: python\n\n    >>> string = """\n    ... def function(param1, param2, param3) -> None:\n    ...     \'\'\'Summary for passing docstring...\n    ...\n    ...     Explanation for passing docstring...\n    ...\n    ...     :param param1: About param1.\n    ...     :param param2: About param2.\n    ...     :param param3: About param3.\n    ...     \'\'\'\n    ...     """\n    >>> docsig(string=string)\n    0\n\n.. code-block:: python\n\n    >>> string = """\n    ... def function(param1, param2) -> None:\n    ...     \'\'\'Summary for failing docstring...\n    ...\n    ...     Explanation for failing docstring...\n    ...\n    ...     :param param1: About param1.\n    ...     :param param2: About param2.\n    ...     :param param3: About param3.\n    ...     \'\'\'\n    ... """\n    >>> docsig(string=string)\n    2\n    -\n    def function(✓param1, ✓param2, ✖None) -> ✓None:\n        """\n        :param param1: ✓\n        :param param2: ✓\n        :param param3: ✖\n        """\n    <BLANKLINE>\n    E102: includes parameters that do not exist\n    <BLANKLINE>\n    1\n\nA full list of checks can be found `here <https://docsig.readthedocs.io/en/latest/docsig.html#docsig-messages>`_\n\nClasses\n#######\nChecking a class docstring is not enabled by default, as this check is opinionated, and won\'t suite everyone\n\nThis check will check documentation of `__init__` under the class docstring, and not under `__init__` itself\n\n.. code-block:: python\n\n    >>> string = """\n    ... class Klass:\n    ...     \'\'\'Summary for failing docstring...\n    ...\n    ...     Explanation for failing docstring...\n    ...\n    ...     :param param1: About param1.\n    ...     :param param2: About param2.\n    ...     :param param3: About param3.\n    ...     \'\'\'\n    ...     def __init__(self, param1, param2) -> None:\n    ...         pass\n    ... """\n    >>> docsig(string=string, check_class=True)\n    11 in Klass\n    -----------\n    class Klass:\n        """\n        :param param1: ✓\n        :param param2: ✓\n        :param param3: ✖\n        """\n    <BLANKLINE>\n        def __init__(✓param1, ✓param2, ✖None) -> ✓None:\n    <BLANKLINE>\n    E102: includes parameters that do not exist\n    <BLANKLINE>\n    1\n\nChecking class docstrings can be permanently enabled in the pyproject.toml file\n\n.. code-block:: toml\n\n    [tool.docsig]\n    check-class = true\n',
    'author': 'jshwi',
    'author_email': 'stephen@jshwisolutions.com',
    'maintainer': 'jshwi',
    'maintainer_email': 'stephen@jshwisolutions.com',
    'url': 'https://pypi.org/project/docsig/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
docsig._core
============
"""
from __future__ import annotations

from pathlib import Path as _Path

from ._display import Display as _Display
from ._display import Failure as _Failure
from ._display import Failures as _Failures
from ._display import FuncStr as _FuncStr
from ._module import Modules as _Modules
from ._module import Parent as _Parent
from ._report import generate_report as _generate_report


def _run_check(  # pylint: disable=too-many-arguments
    parent: _Parent,
    check_class: bool,
    check_dunders: bool,
    check_overridden: bool,
    check_protected: bool,
    no_ansi: bool,
    targets: list[str],
    disable: list[str],
) -> _Failures:
    failures = _Failures()
    for func in parent:
        if not (func.isoverridden and not check_overridden) and (
            not (func.isprotected and not check_protected)
            and not (func.isinit and not check_class)
            and not (func.isdunder and not check_dunders)
        ):
            report = _generate_report(func, targets, disable)
            if report:
                failures.append(
                    _Failure(func, _FuncStr(func, no_ansi), report)
                )

    return failures


def docsig(
    *path: _Path,
    string: str | None = None,
    check_class: bool = False,
    check_dunders: bool = False,
    check_overridden: bool = False,
    check_protected: bool = False,
    no_ansi: bool = False,
    summary: bool = False,
    targets: list[str] | None = None,
    disable: list[str] | None = None,
) -> int:
    """Package's core functionality.

    Populate a sequence of module objects before iterating over their
    top-level functions and classes.

    If any of the functions within the module - and methods within its
    classes - fail, print the resulting function string representation
    and report.

    :param path: Path(s) to check.
    :param string: String to check.
    :param check_class: Check class docstrings.
    :param check_dunders: Check dunder methods
    :param check_overridden: Check overridden methods
    :param check_protected: Check protected functions and classes.
    :param no_ansi: Disable ANSI output.
    :param summary: Print a summarised report.
    :param targets: List of errors to target.
    :param disable: List of errors to disable.
    :return: Exit status for whether test failed or not.
    """
    modules = _Modules(*path, string=string)
    display = _Display(no_ansi)
    for module in modules:
        for top_level in module:
            if not top_level.isprotected or check_protected:
                failures = _run_check(
                    top_level,
                    check_class,
                    check_dunders,
                    check_overridden,
                    check_protected,
                    no_ansi,
                    targets or [],
                    disable or [],
                )
                if failures:
                    display[top_level.path].append(failures)

    if summary:
        display.summarise()
    else:
        display.report()

    return int(bool(display))



"""
docsig._module
==============
"""
from __future__ import annotations

from pathlib import Path as _Path

import astroid as _ast

from ._function import Function as _Function
from ._objects import MutableSequence as _MutableSequence
from ._utils import isprotected as _isprotected


class Parent(_MutableSequence[_Function]):
    """Represents an object that contains functions or methods.

    :param node: Parent's abstract syntax tree.
    :param path: Path to base path representation on.
    """

    def __init__(
        self, node: _ast.Module | _ast.ClassDef, path: _Path | None = None
    ) -> None:
        super().__init__()
        self._name = node.name
        self._path = f"{path}:" if path is not None else ""
        for subnode in node.body:
            if isinstance(subnode, _ast.FunctionDef):
                self.append(_Function(subnode))

    @property
    def path(self) -> str:
        """Representation of path to parent."""
        return self._path

    @property
    def isprotected(self) -> bool:
        """Boolean value for whether class is protected."""
        return _isprotected(self._name)


class _Module(_MutableSequence[Parent]):
    """Represents a module with top level functions and classes.

    :param node: Module's abstract syntax tree.
    :param path: Path for naming module and classes.
    """

    def __init__(self, node: _ast.Module, path: _Path | None = None) -> None:
        super().__init__()
        self.append(Parent(node, path))
        for subnode in node.body:
            if isinstance(subnode, _ast.ClassDef):
                self.append(Parent(subnode, path))


class Modules(_MutableSequence[_Module]):
    """Sequence of ``Module`` objects parsed from Python modules or str.

    Recursively collect Python files from within all dirs that exist
    under paths provided.

    If string is provided, ignore paths.

    :param paths: Path(s) to pase ``Module``(s) from.
    :param string: String to parse if provided.
    """

    def __init__(self, *paths: _Path, string: str | None = None) -> None:
        super().__init__()
        if string is not None:
            self.append(_Module(_ast.parse(string)))
        else:
            for path in paths:
                self._populate(path)

    def _populate(self, root: _Path) -> None:
        if root.is_file() and root.name.endswith(".py"):
            self.append(_Module(_ast.parse(root.read_text()), root))

        if root.is_dir():
            for path in root.iterdir():
                self._populate(path)



"""
docsig._utils
=============
"""
from __future__ import annotations

from difflib import SequenceMatcher as _SequenceMatcher


def isprotected(name: str | None) -> bool:
    """Confirm whether attribute is protected or not.

    :param name: Name to check.
    :return: Boolean value for whether attribute is protected or not.
    """
    return str(name).startswith("_")


def almost_equal(str1: str, str2: str, mini: float, maxi: float) -> bool:
    """Show result for more than the minimum but less than the maximum.

    :param str1: String one to compare with string two.
    :param str2: String two to compare with string one.
    :param mini: Minimum difference allowed between two strings.
    :param maxi: Maximum difference allowed to be considered almost
        equal.
    :return: Boolean result for whether both strings are almost equal.
    """
    return mini < _SequenceMatcher(a=str1, b=str2).ratio() < maxi



"""
docsig._config
==============
"""
from argparse import HelpFormatter as _HelpFormatter
from pathlib import Path as _Path

from arcon import ArgumentParser as _ArgumentParser

from ._display import color as _color
from ._version import __version__

NAME = __name__.split(".", maxsplit=1)[0]


class Parser(_ArgumentParser):
    """Parse commandline arguments."""

    def __init__(self) -> None:
        super().__init__(
            version=__version__,
            prog=_color.cyan.get(NAME),
            formatter_class=lambda prog: _HelpFormatter(
                prog, max_help_position=45
            ),
            description="Check signature params for proper documentation",
        )
        self._add_arguments()
        self.args = self.parse_args()

    def _add_arguments(self) -> None:
        self.add_argument(
            "path",
            nargs="*",
            action="store",
            type=_Path,
            default=[_Path(".")],
            help="directories or files to check (default: .)",
        )
        self.add_argument(
            "-c",
            "--check-class",
            action="store_true",
            help="check class docstrings",
        )
        self.add_argument(
            "-D",
            "--check-dunders",
            action="store_true",
            help="check dunder methods",
        )
        self.add_argument(
            "-o",
            "--check-overridden",
            action="store_true",
            help="check overridden methods",
        )
        self.add_argument(
            "-p",
            "--check-protected",
            action="store_true",
            help="check protected functions and classes",
        )
        self.add_argument(
            "-n", "--no-ansi", action="store_true", help="disable ansi output"
        )
        self.add_argument(
            "-S",
            "--summary",
            action="store_true",
            help="print a summarised report",
        )
        self.add_argument(
            "-s",
            "--string",
            action="store",
            metavar="STR",
            help="string to parse instead of files",
        )
        self.add_list_argument(
            "-d",
            "--disable",
            metavar="LIST",
            help="comma separated list of rules to disable",
        )
        self.add_list_argument(
            "-t",
            "--target",
            metavar="LIST",
            help="comma separated list of rules to target",
        )



"""
docsig._objects
===============
"""
from __future__ import annotations

import typing as _t

T = _t.TypeVar("T")
KT = _t.TypeVar("KT")
VT = _t.TypeVar("VT")


class MutableSequence(_t.MutableSequence[T]):
    """List-object to inherit from.

    :param iterable: Initial iterable to to construct sequence with.
    """

    def __init__(self, iterable: _t.Iterable[T] = ()) -> None:
        self._list: list[T] = []
        self.extend(iterable)

    def insert(self, index: int, value: T) -> None:
        self._list.insert(index, value)

    @_t.overload
    def __getitem__(self, i: int) -> T:
        ...

    @_t.overload
    def __getitem__(self, s: slice) -> _t.MutableSequence[T]:
        ...

    def __getitem__(self, i):
        return self._list.__getitem__(i)

    @_t.overload
    def __setitem__(self, i: int, o: T) -> None:
        ...

    @_t.overload
    def __setitem__(self, s: slice, o: _t.Iterable[T]) -> None:
        ...

    def __setitem__(self, i, o):
        return self._list.__setitem__(i, o)

    @_t.overload
    def __delitem__(self, i: int) -> None:
        ...

    @_t.overload
    def __delitem__(self, i: slice) -> None:
        ...

    def __delitem__(self, i):
        return self._list.__delitem__(i)

    def __len__(self):
        return self._list.__len__()


class MutableMapping(_t.MutableMapping[KT, VT]):
    """Dict-object to inherit from."""

    def __init__(self) -> None:
        self._dict: dict[KT, VT] = {}

    def __setitem__(self, __k: KT, __v: VT) -> None:
        self._dict.__setitem__(__k, __v)

    def __delitem__(self, __v: KT) -> None:
        self._dict.__delitem__(__v)

    def __getitem__(self, __k: KT) -> VT:
        return self._dict.__getitem__(__k)

    def __len__(self) -> int:
        return self._dict.__len__()

    def __iter__(self) -> _t.Iterator[KT]:
        return self._dict.__iter__()
